@@ -4,8 +4,8 @@
 ;;
 ;; Author: Wilfred Hughes <me@wilfred.me.uk>
 ;; Created: 11 January 2013
-;; Version: 0.44
-
+;; Version: 0.45
+;; Package-Requires: ((dash "2.8.0") (s "1.9.0"))
 ;;; Commentary:
 
 ;; Please see README.md for documentation, or read it online at
@@ -34,6 +34,9 @@
 ;;; Code:
 (eval-when-compile (require 'cl)) ;; dolist, defun*, flet
 (require 'dired) ;; dired-sort-inhibit
+(require 'dash)
+(require 's)
+(require 'ido)  ;; completion
 
 (defcustom ag-executable
   "ag"
@@ -85,6 +88,11 @@ If set to nil, fall back to finding VCS root directories."
                  (function :tag "Function"))
   :group 'ag)
 
+(defcustom ag-ignore-list nil
+  "A list of patterns to ignore when searching."
+  :type '(repeat (string))
+  :group 'ag)
+
 (require 'compile)
 
 ;; Although ag results aren't exactly errors, we treat them as errors
@@ -101,7 +109,7 @@ If set to nil, fall back to finding VCS root directories."
 
 (defun ag/next-error-function (n &optional reset)
   "Open the search result at point in the current window or a
-different window, according to `ag-open-in-other-window'."
+different window, according to `ag-reuse-window'."
   (if ag-reuse-window
       ;; prevent changing the window
       (flet ((pop-to-buffer (buffer &rest args)
@@ -110,22 +118,28 @@ different window, according to `ag-open-in-other-window'."
     ;; just navigate to the results as normal
     (compilation-next-error-function n reset)))
 
+;; Note that we want to use as tight a regexp as we can to try and
+;; handle weird file names (with colons in them) as well as possible.
+;; E.g. we use [1-9][0-9]* rather than [0-9]+ so as to accept ":034:"
+;; in file names.
+(defvar ag/file-column-pattern
+  "^\\(.+?\\):\\([1-9][0-9]*\\):\\([1-9][0-9]*\\):"
+  "A regexp pattern that groups output into filename, line number and column number.")
+
 (define-compilation-mode ag-mode "Ag"
   "Ag results compilation mode"
-  (let ((smbl  'compilation-ag-nogroup)
-        ;; Note that we want to use as tight a regexp as we can to try and
-        ;; handle weird file names (with colons in them) as well as possible.
-        ;; E.g. we use [1-9][0-9]* rather than [0-9]+ so as to accept ":034:"
-        ;; in file names.
-        (pttrn '("^\\([^:\n]+?\\):\\([1-9][0-9]*\\):\\([1-9][0-9]*\\):" 1 2 3)))
-    (set (make-local-variable 'compilation-error-regexp-alist) (list smbl))
-    (set (make-local-variable 'compilation-error-regexp-alist-alist) (list (cons smbl pttrn))))
+  (set (make-local-variable 'compilation-error-regexp-alist)
+       (list 'compilation-ag-nogroup))
+  (set (make-local-variable 'compilation-error-regexp-alist-alist)
+       (list (cons 'compilation-ag-nogroup (list ag/file-column-pattern 1 2 3))))
   (set (make-local-variable 'compilation-error-face) 'ag-hit-face)
   (set (make-local-variable 'next-error-function) 'ag/next-error-function)
   (add-hook 'compilation-filter-hook 'ag-filter nil t))
 
 (define-key ag-mode-map (kbd "p") 'compilation-previous-error)
 (define-key ag-mode-map (kbd "n") 'compilation-next-error)
+(define-key ag-mode-map (kbd "k") '(lambda () (interactive) 
+                                     (let (kill-buffer-query-functions) (kill-buffer))))
 
 (defun ag/buffer-name (search-string directory regexp)
   "Return a buffer name formatted according to ag.el conventions."
@@ -134,8 +148,13 @@ different window, according to `ag-open-in-other-window'."
    (regexp (format "*ag search regexp:%s dir:%s*" search-string directory))
    (:else (format "*ag search text:%s dir:%s*" search-string directory))))
 
+(defun ag/format-ignore (ignores)
+  "Prepend '--ignore' to every item in IGNORES."
+  (apply 'append
+         (mapcar (lambda (item) (list "--ignore" item)) ignores)))
+
 (defun* ag/search (string directory
-                          &key (regexp nil) (file-regex nil))
+                          &key (regexp nil) (file-regex nil) (file-type nil))
   "Run ag searching for the STRING given in DIRECTORY.
 If REGEXP is non-nil, treat STRING as a regular expression."
   (let ((default-directory (file-name-as-directory directory))
@@ -148,6 +167,10 @@ If REGEXP is non-nil, treat STRING as a regular expression."
       (setq arguments (append '("--nocolor") arguments)))
     (when (char-or-string-p file-regex)
       (setq arguments (append `("--file-search-regex" ,file-regex) arguments)))
+    (when file-type
+      (setq arguments (cons (format "--%s" file-type) arguments)))
+    (when ag-ignore-list
+      (setq arguments (append (ag/format-ignore ag-ignore-list) arguments)))
     (unless (file-exists-p default-directory)
       (error "No such directory %s" default-directory))
     (let ((command-string
@@ -185,7 +208,7 @@ a PCRE pattern that matches files with that extension.
 Returns an empty string otherwise."
   (let ((file-name (buffer-file-name)))
     (if (stringp file-name)
-        (format "\\.%s" (file-name-extension file-name))
+        (format "\\.%s$" (ag/escape-pcre (file-name-extension file-name)))
       "")))
 
 (defun ag/longest-string (&rest strings)
@@ -337,15 +360,16 @@ If called with a prefix, prompts for flags to pass to ag."
    (ag/search string directory))
 
 ;;;###autoload
-(defun ag-files (string file-regex directory)
-  "Search using ag in a given DIRECTORY and file type regex FILE-REGEX
-for a given search STRING, with STRING defaulting to the symbol under point.
+(defun ag-files (string file-type directory)
+  "Search using ag in a given DIRECTORY for a given search STRING,
+limited to files that match FILE-TYPE. STRING defaults to
+the symbol under point.
 
 If called with a prefix, prompts for flags to pass to ag."
   (interactive (list (read-from-minibuffer "Search string: " (ag/dwim-at-point))
-                     (read-from-minibuffer "In filenames matching PCRE: " (ag/buffer-extension-regex))
+                     (ag/read-file-type)
                      (read-directory-name "Directory: ")))
-  (ag/search string directory :file-regex file-regex))
+  (apply 'ag/search string directory file-type))
 
 ;;;###autoload
 (defun ag-regexp (string directory)
@@ -366,14 +390,15 @@ If called with a prefix, prompts for flags to pass to ag."
   (ag/search string (ag/project-root default-directory)))
 
 ;;;###autoload
-(defun ag-project-files (string file-regex)
-  "Search using ag in a given DIRECTORY and file type regex FILE-REGEX
-for a given search STRING, with STRING defaulting to the symbol under point.
+(defun ag-project-files (string file-type)
+  "Search using ag for a given search STRING,
+limited to files that match FILE-TYPE. STRING defaults to the
+symbol under point.
 
 If called with a prefix, prompts for flags to pass to ag."
   (interactive (list (read-from-minibuffer "Search string: " (ag/dwim-at-point))
-                     (read-from-minibuffer "In filenames matching PCRE: " (ag/buffer-extension-regex))))
-  (ag/search string (ag/project-root default-directory) :file-regex file-regex))
+                     (ag/read-file-type)))
+  (apply 'ag/search string (ag/project-root default-directory) file-type))
 
 ;;;###autoload
 (defun ag-project-regexp (regexp)
@@ -428,7 +453,7 @@ See also `find-dired'."
          (buffer-name (if ag-reuse-buffers
                           "*ag dired*"
                         (format "*ag dired pattern:%s dir:%s*" regexp dir)))
-         (cmd (concat "ag --nocolor -g '" regexp "' "
+         (cmd (concat ag-executable " --nocolor -g '" regexp "' "
                       (shell-quote-argument dir)
                       " | grep -v '^$' | sed s/\\'/\\\\\\\\\\'/ | xargs -I '{}' ls "
                       dired-listing-switches " '{}' &")))
@@ -453,7 +478,7 @@ See also `find-dired'."
       (set (make-local-variable 'dired-sort-inhibit) t)
       (set (make-local-variable 'revert-buffer-function)
            `(lambda (ignore-auto noconfirm)
-              (ag-dired ,orig-dir ,regexp)))
+              (ag-dired-regexp ,orig-dir ,regexp)))
       (if (fboundp 'dired-simple-subdir-alist)
           (dired-simple-subdir-alist)
         (set (make-local-variable 'dired-subdir-alist)
@@ -525,5 +550,28 @@ This function is called from `compilation-filter-hook'."
           (while (re-search-forward "\033\\[[0-9;]*[mK]" end 1)
             (replace-match "" t t)))))))
 
+(defun ag/get-supported-types ()
+  "Query the ag executable for which file types it recognises."
+  (let* ((ag-output (shell-command-to-string (format "%s --list-file-types" ag-executable)))
+         (lines (-map 's-trim (s-lines ag-output)))
+         (types (--keep (when (s-starts-with? "--" it) (s-chop-prefix "--" it )) lines))
+         (extensions (--map (s-split "  " it) (--filter (s-starts-with? "." it) lines))))
+    (-zip types extensions)))
+
+(defun ag/read-file-type ()
+  "Prompt the user for a known file type, or let them specify a PCRE regex."
+  (let* ((all-types-with-extensions (ag/get-supported-types))
+         (all-types (mapcar 'car all-types-with-extensions))
+         (file-type
+          (ido-completing-read "Select file type: "
+                               (append '("custom (provide a PCRE regex)") all-types)))
+         (file-type-extensions
+          (cdr (assoc file-type all-types-with-extensions))))
+    (if file-type-extensions
+        (list :file-type file-type)
+      (list :file-regex
+            (read-from-minibuffer "Filenames which match PCRE: "
+                                  (ag/buffer-extension-regex))))))
+
 (provide 'ag)
 ;;; ag.el ends here
